'''
bwobject -- Base class required to use most bullwinkle features

This module exports the BWObject base class that integrates the necessary
metaclass that make most of the other features possible.
'''

from __version__ import *
import sys

class BWObjectMeta(type):
    '''
    Provides the machinery for making Object work.  It scans any derived
    class of Objecct for __bindclass__ and, if found, calls it with the
    class and attribute name as parameters.  If the function returns
    anything but None, that is used to replace the called function.

    See also:
    Object.makemeta()
    '''

    def __init__(cls, typename, typebases, typedict):
        super(BWObjectMeta, cls).__init__(typename, typebases, typedict)
        for name, value in typedict.iteritems():
            fn = getattr(value, '__bindclass__', None)
            if fn is not None:
                replacement = fn(cls, name)
                if replacement is not None:
                    if replacement is type(None):
                        delattr(cls, name)
                    else:
                        setattr(cls, name, replacement)

class BWObject(object):
    '''
    Base class for any object that wants to use bullwinkle extensions to
    object.  These extensions include:

    * Ability to use super_* method decorators to autmatically call
        superclass methods.

    * Ability for member objects to bind to the class automatically
        by providing a __bindclass__ method on the object.

    * Ability to define complex properties by simple declaration.
    '''

    __metaclass__ = BWObjectMeta
    __positional__ = ()

    def __init__(_self, *_args, **_kw):
        positional = _self.__positional__
        if len(_args) > len(positional):
            raise TypeError(
                '%s() can only accept up to %d positional arguments.'
                % (type(_self).__name__, len(positional)))
        elif positional:
            for name, value in zip(positional, _args):
                if name in _kw:
                    raise TypeError('Multiple definitions for %r' % name)
                else:
                    _kw[name] = value
        cls = type(_self)
        for name, value in _kw.iteritems():
            obj = getattr(cls, name, None)
            fn = getattr(obj, '__initobj__', None)
            if fn is not None:
                fn(_self, name, value)
        required = getattr(_self, '__required__', ())
        if required:
            NOT_FOUND = type(None)
            missing = []
            for name in required:
                if getattr(_self, name, NOT_FOUND) is NOT_FOUND:
                    missing.append(name)
            if missing:
                raise TypeError('%s need to be specified when constructing %s.'
                                % (', '.join(missing), type(_self).__name__))

    @classmethod
    def mix(cls, *others, **kw):
        '''
        When mixing in Object with other classes, this class method will
        make creating the correct metaclass easier.

        For example, consider if we previously have this class with a
        metaclass:
        >>> class MyClassMeta(type):
        ...     def __init__(cls, *args):
        ...         cls.MyClassMeta = 'HERE'
        >>> class MyClass(object):
        ...     __metaclass__ = MyClassMeta

        Now we want to add Object to the cl;ass hierarchy:
        >>> class MyJoined(MyClass, BWObject):
        ...     pass
        ... #doctest: +ELLIPSIS
        Traceback (most recent call last):
            ...
        TypeError: Error when calling the metaclass bases
            metaclass conflict: ...

        To fix this, we will create a mix base between Object and other
        classes:
        >>> class MyJoined(BWObject.mix(MyClass)):
        ...     pass
        >>> obj = MyJoined()
        >>> obj.MyClassMeta
        'HERE'
        '''

        names = ', '.join(c.__name__ for c in (others + (cls,))),
        basemeta = []
        bases = others + (cls,)
        for c in bases:
            metaclass = getattr(c, '__metaclass__', None)
            if metaclass is not None:
                basemeta.append(metaclass)
        metabase = type('<metaclass for %s>' % names, tuple(basemeta), {})
        return metabase('<mixin-base for %s>' % names, bases,
            dict(kw, __module__=cls.__module__))

    def __str__(self):
        format = getattr(self, '__bwformat__', None)
        if format:
            return format % self.__objdict__
        else:
            return repr(self)

    @property
    def __objdict__(self):
        class ObjectDict(object):
            def __getitem__(od, name):
                return getattr(self, name)
        return ObjectDict()

    def __repr__(self):
        name = type(self).__name__
        members = sorted(getattr(self, '__bwmembers__', ()),
                         lambda a, b: cmp(a.__name__, b.__name__))
        kw = ', '.join('%s=%r' % (member.__name__, member.__get__(self))
                       for member in members)
        return '%s(%s)' % (name, kw)




'''
bwmethods -- Bullwinkle method construction kit.

Provides functions for the construction of methods at compile time.  Many
of these are meant for overriding base class methods and manipulating their
results.  This draws etensively from Perl::Moose but adds some Python
flavour to make things fit in a Pythonic world.
'''

from __version__ import *
from bwobject import BWObject

class MethodBuilder(BWObject):
    require_base = True
    want_args = True
    want_result = False
    can_override = False

    def __new__(cls, fn=None, **_kw):
        if fn is None:
            return lambda f: cls(f, **_kw)
        else:
            return super(MethodBuilder, cls).__new__(cls)

    def __init__(self, fn, **_kw):
        self.fn = fn
        self.init(**_kw)

    def init(self, require_base=None, want_args=None,
                   can_override=None, want_result=None):
        if require_base is not None:
            self.require_base = require_base
        if want_args is not None:
            self.want_args = want_args
        if want_result is not None:
            self.want_result = want_result
        if can_override is not None:
            self.can_override = can_override

    def __bindclass__(self, cls, name):
        method = self.build_wrapper(cls, name)
        method.__name__ = name
        method.__doc__ = self.fn.__doc__
        return method

    def build_wrapper(self, cls, name):
        code = ['def wrapper(_self, *_args, **_kw):']
        self.encode_wrapper(code)
        v = self.get_wrapper_dict(cls, name)
        src = '\n'.join(code)
        exec src in v
        wrapper = v.pop('wrapper')
        wrapper.__src__ = src
        return wrapper

    def get_wrapper_dict(self, cls, name):
        d = dict(cls=cls, name=name, fn=self.fn)
        if self.can_override:
            d['type_none'] = type(None)
        return d

    def encode_wrapper(self, code):
        raise NotImplementedError('No encode defined for %r' % self)

class SuperMethodBuilder(MethodBuilder):
    '''
    Base class for all super-class overriding methods.
    '''

    def __bindclass__(self, cls, name):
        if self.require_base:
            if getattr(super(cls, cls), name, None) is None:
                raise TypeError('Method %r is required in superclasses of %r' %
                    (name, cls.__name__))
        return super(SuperMethodBuilder, self).__bindclass__(cls, name)

    def encode_get_base_result(self, code):
        if self.require_base:
            code.append(
                '    res = getattr(super(cls, _self), name)(*_args, **_kw)',
            )
        else:
            code.extend((
                '    super_fn = getattr(super(cls, _self), name, None)',
                '    if super_fn is None:',
                '        res = None',
                '    else:',
                '        res = super_fn(*_args, **_kw)'
            ))

    def encode_do_fn(self, code):
        if self.can_override:
            code.append('    override = ' + self.get_fn_call())
        else:
            code.append('    ' + self.get_fn_call())

    def get_fn_call(self):
        args = ('_self',) + self.get_fn_positional_args()
        if self.want_args:
            args += ('*_args',)
        args += self.get_fn_keyword_args()
        if self.want_args:
            args += ('**_kw',)
        return 'fn(%s)' % ', '.join(args)

    def get_fn_positional_args(self):
        if self.want_result:
            return ('res',)
        else:
            return ()

    def get_fn_keyword_args(self):
        return ()

    def encode_override(self, code):
        if self.can_override:
            code.extend((
                '    if override is type_none:',
                '        return None',
                '    elif override is not None:',
                '        return override',
            ))

    def encode_return(self, code):
        code.append('    return res')

class OverrideSuperMethodBuild(SuperMethodBuilder):
    '''
    >>> class Sub(BWObject):
    ...     @override_super
    ...     def fn(self):
    ...         return 'Hello'
    ...
    Traceback (most recent call last):
        ...
    TypeError: Method 'fn' is required in superclasses of 'Sub'
    >>> class Base(BWObject):
    ...     def fn(self):
    ...         return 'Hello'
    ...
    >>> class Sub(Base):
    ...     @override_super
    ...     def fn(self):
    ...         return 'World'
    ...
    >>> b = Base()
    >>> s = Sub()
    >>> b.fn()
    'Hello'
    >>> s.fn()
    'World'
    '''

    def build_wrapper(self, cls, name):
        return self.fn
override_super = OverrideSuperMethodBuild

class BeforeSuperMethodBuilder(SuperMethodBuilder):
    '''
    >>> class Sqrt(BWObject):
    ...     def sqrt(self, arg):
    ...         return arg ** .5
    ...
    >>> class OnlyPositiveSqrt(Sqrt):
    ...     @before_super
    ...     def sqrt(self, arg):
    ...         if (arg < 0):
    ...             return 0
    ...
    >>> s = Sqrt()
    >>> ops = OnlyPositiveSqrt()
    >>> s.sqrt(-1)
    Traceback (most recent call last):
        ...
    ValueError: negative number cannot be raised to a fractional power
    >>> ops.sqrt(-1)
    0
    '''

    can_override = True

    @override_super
    def encode_wrapper(self, code):
        self.encode_do_fn(code)
        self.encode_override(code)
        self.encode_get_base_result(code)
        self.encode_return(code)
before_super = BeforeSuperMethodBuilder

class AfterSuperMethodBuilder(SuperMethodBuilder):
    '''
    >>> class Base(BWObject):
    ...     y = None
    ...     def fn(self, x):
    ...         return x * 2
    ...
    >>> class Sub(Base):
    ...     @after_super
    ...     def fn(self, x):
    ...         self.y = x
    ...         return self.y
    ...
    >>> b = Base()
    >>> s = Sub()
    >>> b.fn(5)
    10
    >>> s.fn(5)
    10
    >>> b.y
    >>> s.y
    5
    '''

    @override_super
    def encode_wrapper(self, code):
        self.encode_get_base_result(code)
        self.encode_do_fn(code)
        self.encode_override(code)
        self.encode_return(code)
after_super = AfterSuperMethodBuilder

class FollowSuperMethodBuilder(AfterSuperMethodBuilder):
    '''
    >>> class Base(BWObject):
    ...     def __init__(self, x):
    ...         self.x = x
    ...
    >>> class Sub(Base):
    ...     @follow_super
    ...     def __init__(self):
    ...         self.x *= 2
    ...
    >>> b = Base(5)
    >>> s = Sub(5)
    >>> b.x
    5
    >>> s.x
    10
    '''
    want_args = False
    can_override = False
follow_super = FollowSuperMethodBuilder

class FilterSuperMethodBuilder(AfterSuperMethodBuilder):
    '''
    >>> class Doubler(BWObject):
    ...     def apply(self, arg):
    ...         return arg * 2
    ...
    >>> class Quadrupler(Doubler):
    ...     @filter_super
    ...     def apply(self, res, arg):
    ...         return res * 2
    ...
    >>> d = Doubler()
    >>> q = Quadrupler()
    >>> d.apply(2)
    4
    >>> q.apply(2)
    8
    '''
    want_result = True
    can_override = True
filter_super = FilterSuperMethodBuilder

class AroundSuperMethodBuilder(SuperMethodBuilder):
    '''
    >>> class Base(BWObject):
    ...     def fn(self, x):
    ...         return x ** 2
    ...
    >>> class Sub(Base):
    ...     @around_super
    ...     def fn(self, fn, x):
    ...         x += 1
    ...         res = fn(x * 2)
    ...         return res - 1
    ...
    >>> b = Base()
    >>> s = Sub()
    >>> b.fn(5)
    25
    >>> s.fn(5)
    143
    '''

    want_result = False
    can_override = False

    @override_super
    def encode_wrapper(self, code):
        self.encode_do_fn(code)
        self.encode_return(code)

    @follow_super
    def init(self):
        if self.want_result:
            raise TypeError('Cannot get base result in around methods')
        if self.can_override:
            raise TypeError('Overrides are meaningless in around methods')

    @override_super
    def encode_do_fn(self, code):
        if self.require_base:
            code.append('    super_fn = getattr(super(cls, _self), name)')
        else:
            code.append('    super_fn = getattr(super(cls, _self), name, None)')
        code.append('    res = ' + self.get_fn_call())

    @filter_super(want_args=False)
    def get_fn_positional_args(self, res):
        return res + ('super_fn',)

around_super = AroundSuperMethodBuilder

def override_result(v, type_none=type(None)):
    return type_none if v is None else v




'''
bwthrowable -- throw and catch objects on the stack

Provides a means to send contextual objects to called frames.  Thrown
variables are stored based on their Python type so that they may be caught
either the type of a subtype.

For example, consider an HTTP request class:
>>> class HTTPRequest(BWThrowable):
...     def __init__(self, method, path):
...         self.method = method
...         self.path = path
...

It is often the case that a web framework will define one and want to send
it along as context.  Often it is inconvenient to send the context as
parameters to all called functions, so:

>>> req = HTTPRequest('GET', '/index.html')
>>> req.throw()

Now any function that needs access to the request but cannot acquire it by
parameter can pick it up using .catch():

>>> def controller():
...     req = HTTPRequest.catch()
...     print req.path
...
>>> controller()
/index.html

The throw part can be skipped via with statements:

>>> with HTTPRequest('GET', '/image/pic.png') as req:
...     controller()
...
/image/pic.png

In addition, individual keys can be thrown and caught, though throw() is
optional if the key name is a simple string:

>>> def catcher(name):
...     print catch(name)
...
>>> throw('x', 5)
>>> catcher('x')
5
>>> x = 7
>>> catcher('x')
7

A special singleton global, "TC." is available to streamline this further.
Anything assiged to TC, by attribute or item, is thrown().  Anything
accessed via TC, by attribute or item, is catch()'ed:

>>> def catcher(name):
...     print TC[name]
...
>>> TC.x = 5
>>> catcher('x')
5
'''

from __version__ import *
import sys

def throw(key, value, frame=0):
    sys._getframe(frame + 1).f_locals[key] = value

def catch(key, default=None, NOT_FOUND=KeyError):
    f = sys._getframe(1)
    while f is not None:
        obj = f.f_locals.get(key, NOT_FOUND)
        if obj is not NOT_FOUND:
            return obj
        f = f.f_back
    else:
        return default

class BWThrowableNotFoundError(Exception):
    def __init__(self, cls):
        self.cls = cls

class BWThrowable(object):
    def throw(self, frame=0):
        sys._getframe(frame + 1).f_locals[type(self)] = self
    __enter__ = throw

    @classmethod
    def catch(cls):
        top = sys._getframe(1)
        for base in cls.__mro__:
            f = top
            while f is not None:
                obj = f.f_locals.get(base)
                if obj is not None:
                    return obj
                else:
                    f = f.f_back
        else:
            return self.uncaught()

    @classmethod
    def uncaught(cls):
        return cls()

    class DefaultNone(object):
        @classmethod
        def uncaught(self):
            return None

    class DefaultError(object):
        error_type = BWThrowableNotFoundError

        @classmethod
        def uncaught(cls):
            raise error_type(cls)

    def __exit__(self, *_args):
        pass

class BWThrowCatch(object):
    def __getitem__(self, key):
        if isinstance(key, type):
            return key.catch()
        else:
            return catch(key)

    def __setitem__(self, key, value):
        throw(key, value, 1)

    def __getattr__(self, name):
        return catch(name)

    def __setattr__(self, name, value):
        throw(name, value, 1)

TC = BWThrowCatch()




'''
__license__ -- bullwinkle license module

Loads and dumps the included LICENSE file if used via python -m or simply
exports the entire file as a LICENSE module attribute.
'''

from __version__ import *
import os

LICENSE = open(os.path.join(os.path.dirname(__file__), 'LICENSE')).read()

if __name__ == '__main__':
    print LICENSE




'''
bwversion -- Manage version numbers and change logs

This module simplifies the process of specifying and comparing version
descriptors and then grouping them into change log objects.  The objects
can then be printed easily.
'''

class Version(tuple):
    '''
    >>> Version(0.1)
    0.1
    >>> Version('0.1.2')
    0.1.2
    >>> Version('0.1.2') > Version('0.1')
    True
    '''

    def __new__(cls, src, *changes):
        if src is None or isinstance(src, Version):
            return src
        elif isinstance(src, basestring):
            return Version(src.split('.'), *changes)
        elif isinstance(src, (int, float)):
            return Version(str(src), *changes)
        else:
            obj = super(Version, cls).__new__(cls, src)
            obj.changes = changes
            if len(changes) == 0:
                obj.label = None
            elif len(changes) == 1:
                obj.label = changes[0]
            else:
                obj.label = changes
            return obj

    def __add__(self, other):
        if isinstance(other, basestring):
            return str(self) + other
        else:
            return super(Version, self).__add__(other)

    def __radd__(self, other):
        if isinstance(other, basestring):
            return other + str(self)
        else:
            return other.__add__(self)

    def __str__(self):
        return '.'.join(self)
    __repr__ = __str__

class ChangeLog(tuple):
    '''
    >>> ChangeLog(
    ...     Version(0.1, 'Version 0.1'),
    ...     Version(0.2, 'Change A in 0.2', 'Change B in 0.2'),
    ... )
    ...
    0.2
        * Change A in 0.2
        * Change B in 0.2
    <BLANKLINE>
    0.1
        * Version 0.1
    '''

    def __new__(cls, *items):
        return super(ChangeLog, cls).__new__(cls, items)

    def __str__(self):
        return '\n\n'.join(
            '%s\n    * %s' % (
                version,
                '<No log provided>' if version.label is None
                    else version.label if len(version.changes) == 1
                    else '\n    * '.join(map(str, version.changes))
                ) for version in reversed(sorted(self)))

    __repr__ = __str__





'''
__version__ file for bullwinkle.

Contains the version information attributes that are imported by the
packages of this library.  If run with python -m, prints the version
information in SHELL loadable foramt.
'''

from __changelog__ import CHANGELOG

LICENSE = 'LGPL-3.0'
AUTHOR = 'Rich Harkins'
AUTHOR_EMAIL = 'rich.harkins@gmail.com'
VERSION = CHANGELOG[-1]
PROJECT = 'bullwinkle'
LICENSE_INFO = 'Run python -m bullwinkle.LICENSE for license text'
CHANGELOG_INFO = 'Run python -m bullwinkle.__changelog__ for full log'

__author__ = AUTHOR
__version__ = VERSION
__all__ = ['LICENSE', 'AUTHOR', 'AUTHOR_EMAIL',
           'VERSION', 'PROJECT', 'LICENSE_INFO',
           'CHANGELOG_INFO', '__author__', '__version__']

if __name__ == '__main__':
    shvar = PROJECT.replace('-', '_').replace(' ', '_').lower()
    print('')
    print('# %r version information (*SH-compatible format)' % PROJECT)
    print('%s_AUTHOR="%s <%s>"' % (shvar, AUTHOR, AUTHOR_EMAIL))
    print('%s_VERSION=%r' % (shvar, VERSION))
    print('%s_LICENSE=%r' % (shvar, LICENSE))
    print('')
    print('# ' + LICENSE_INFO)
    print('# ' + CHANGELOG_INFO)
    print('')




'''
__test__ -- provides testing for bullwinkle

This can be invoked via python bullwinkle.__test__
'''

import doctest

if __name__ == '__main__':
    import bwversion, bwobject, bwmethod
    import bwcached, bwmember, bwthrowable
    doctest.testmod(bwversion)
    doctest.testmod(bwobject)
    doctest.testmod(bwmethod)
    doctest.testmod(bwcached)
    doctest.testmod(bwmember)
    doctest.testmod(bwthrowable)




'''
bullwinkle -- A Python implementation of some of the best Perl::Moose features

Provides numerous capabilities that originally stemmed from Perl::Moose but
may expand outside of that realm over time.  Among these feeatures are:

 * Superclass method override decorators
 * Simplified type-safe proprety declaration
 * Type-safe __init__ keyword arguments via property attributes

In addition, bullwinkle provides:

 * Automatic class member binding via __bindclass__ methods.
'''

from __version__ import *
from bwobject import BWObject
from bwmethod import (before_super, after_super, follow_super, filter_super,
                      override_super, around_super, override_result)
from bwcached import cached, classcached, cachedmethod, classcachedmethod
from bwmember import member
from bwthrowable import throw, catch, BWThrowable, TC




'''
bwcached -- Lazy loading properties

Provides simple ways to produce lazy loaded values.  These are slimmer than
the assocaited lazy member() functions of bwmember and work for many cases.
In addition, these do NOT require BWObject to work so may be preffed in
some applications.
'''

from __version__ import *
import types

class Volatile(object):
    __slots__ = ['obj']

    def __init__(self, obj):
        self.obj = obj

def cached(fn, Volailte=Volatile):
    '''
    >>> class MyClass(object):
    ...     def __init__(self, x):
    ...         self.x = x
    ...
    ...     @cached
    ...     def doubled(self):
    ...         print "HERE"
    ...         return self.x * 2
    ...
    >>> c = MyClass(10)
    >>> c.doubled
    HERE
    20
    >>> c.doubled
    20
    >>> c.__dict__['doubled']
    20
    '''

    name = fn.__name__
    def wrapper(self, target, cls=None):
        if target is None:
            return fn
        else:
            obj = fn(target)
            if type(obj) is not Volatile:
                target.__dict__[name] = obj
                return obj
            else:
                return obj.obj
    cls = type(fn.__name__,
               (object,),
               dict(__doc__=fn.__doc__, __get__=wrapper))
    return cls()
cached.volatile = Volatile

def classcached(fn, Volailte=Volatile):
    '''
    >>> class MyClass(object):
    ...     x = 10
    ...
    ...     @classcached
    ...     def doubled(self):
    ...         print "HERE"
    ...         return self.x * 2
    ...
    >>> c = MyClass()
    >>> c.doubled
    HERE
    20
    >>> c.doubled
    20
    >>> MyClass.__dict__['doubled']
    20
    '''

    name = fn.__name__
    def wrapper(self, target, cls=None):
        target = cls or type(target)
        obj = fn(target)
        if type(obj) is not Volatile:
            setattr(target, name, obj)
            return obj
        else:
            return obj.obj
    cls = type(fn.__name__,
               (object,),
               dict(__doc__=fn.__doc__, __get__=wrapper))
    return cls()
classcached.volatile = Volatile

def cachedmethod(fn, Volailte=Volatile):
    '''
    >>> class MyClass(object):
    ...     def __init__(self, x):
    ...         self.x = x
    ...
    ...     @cached
    ...     def doubled(self):
    ...         print "HERE"
    ...         return self.x * 2
    ...
    >>> c = MyClass(10)
    >>> c.doubled
    HERE
    20
    >>> c.doubled
    20
    >>> c.__dict__['doubled']
    20
    '''

    name = fn.__name__
    def wrapper(self, target, cls=None, MethodType=types.MethodType):
        if target is None:
            return fn
        else:
            obj = fn(target)
            if type(obj) is not Volatile:
                obj = MethodType(obj, target, type(target))
                target.__dict__[name] = obj
                return obj
            else:
                return obj.obj
    cls = type(fn.__name__,
               (object,),
               dict(__doc__=fn.__doc__, __get__=wrapper))
    return cls()
cached.volatile = Volatile

def classcachedmethod(fn, Volailte=Volatile):
    '''
    >>> class MyClass(object):
    ...     x = 10
    ...
    ...     @classcached
    ...     def doubled(self):
    ...         print "HERE"
    ...         return self.x * 2
    ...
    >>> c = MyClass()
    >>> c.doubled
    HERE
    20
    >>> c.doubled
    20
    >>> MyClass.__dict__['doubled']
    20
    '''

    name = fn.__name__
    def wrapper(self, target, cls=None):
        target = cls or type(target)
        obj = fn(target)
        if type(obj) is not Volatile:
            obj = classmethod(obj)
            setattr(target, name, obj)
            return obj
        else:
            return obj.obj
    cls = type(fn.__name__,
               (object,),
               dict(__doc__=fn.__doc__, __get__=wrapper))
    return cls()
classcached.volatile = Volatile




'''
bwmember -- Easy to build type-safe propreties

=====================
=== Type checking ===
=====================

Python's approach to object creation works very well for the most part.
However, sometimes one wants to build a class that does type checking on
one or more of its members.  In these cases, member objects come into play:

>>> class Point(BWObject):
...     x = member(int)
...     y = member(int)
...     __bwformat__ = '<%(x)d, %(y)d>'
...

Now we can define a point using keyword arguments:

>>> p = Point(x=5, y=7)
>>> p
Point(x=5, y=7)
>>> print p
<5, 7>

Note how the __repr__ came for free and the __str__ was simple to set up.
In addition, we cannot set the values to something invalid:

>>> p.x = 'Hello'
Traceback (most recent call last):
    ...
TypeError: x ('Hello') must be one of: (<type 'int'>)

Now if multiple types are allowed, then they can all be specified as peers.
Each entry can be one of:

 * A python type, in which case the value must be an instance
 * A callable, in which case the value and member object are provided as
    arguments.  The callable returns true if acceptable.
 * A value, in which case if equal to the object is acceptable (None
    works well here).

>>> class Car(BWObject):
...     color = member('BLUE', 'RED', 'BLACK')
...
>>> Car(color='BLUE')
Car(color='BLUE')
>>> Car(color='BLACK')
Car(color='BLACK')
>>> Car(color='WHITE')
Traceback (most recent call last):
    ...
TypeError: color ('WHITE') must be one of: ('BLUE', 'RED', 'BLACK')

=======================
=== Type conversion ===
=======================

The convert() function indicates that a given type can be converted into
from other types.  For example, let's revisit the Point class:

>>> class NewPoint(BWObject):
...     x = member(into(float, int))
...     y = member(into(float, int))
...
>>> NewPoint(x=1.5, y=1.5)
NewPoint(x=1.5, y=1.5)
>>> NewPoint(x=1, y=2)
NewPoint(x=1.0, y=2.0)

Now Point will accept int or float but won't try to convert unless it
receives an int (but not a string).  If only one positional argument is
provided to into(), then any type will be converted using the converter
function.  Conversion will not occur if the object is already of the type
specified.

======================
=== Default values ===
======================

When not provided otherwise, the default attribute of the member (set by
the member constructor) will determine what to do (in this order):

 * If the default memebr is a subclass of Exception it will be called with
    the offending name.

 * If the default member is a type, it is called with no arguments.

 * If the default member is a callable, it will be called with the "self"
    object, the member object, and the name.

 * Otherwise, the default object is used as the default value.

In addition, a builder method name may be specified during member
construction.  If provided, the builder is a name of a method on the object
that will accept the default, possibly use it, and return a built value as
appropriate.

>>> import hashlib
>>> class DBConnection(BWObject):
...     username = member(str, default='nobody')
...     password = member(str, default='nobody', builder='makepw')
...
...     def makepw(self, pw):
...         return hashlib.sha1(pw).digest()
...
>>> conn = DBConnection()
>>> conn.username
'nobody'
>>> print conn.password
6^\xc1zg_2s\xbc\x16\xc7Ga\xad\x83\xf2\xcf\x07\xc5\x9a

=========================
=== Read-Only Members ===
=========================

Members created using ro=True will reject changes after the initial
creation of the object or via lazy loading using default/builder.

>>> class StringJoiner(BWObject):
...     inputs = member(list, tuple)
...     separator = member(str, default=', ')
...     contents = member(str, builder='combine', ro=True)
...     __bwformat__ = '%(contents)s'
...
...     def combine(self, defvalue):
...         return self.separator.join(self.inputs)
...
>>> joiner = StringJoiner(inputs=('Hello', 'world'), separator=' ')
>>> print joiner
Hello world
>>> joiner.contents = 'blah'
Traceback (most recent call last):
    ...
AttributeError: can't set attribute

========================
=== Optional members ===
========================

Members that are not required during definition can be specified with
optional=True.  These memebers will not be checked for value during
object __init__ in BWObject.  Objects with default or builder values will
be set to optional by default.  Setting optional to False will cause the
value to be determined on object init instead of first member use.  Members
are not optional by default otherwise.  

IMPORTANT: Optional members that do not have default/builder settings can
            still generate exceptions when probed.  Generally, default
            and/or builder should be used in lieu of optional=True.

>>> p = Point()
Traceback (most recent call last):
    ...
TypeError: x, y need to be specified when constructing Point.
'''

from __version__ import *
from bwobject import BWObject
from bwmethod import after_super
from bwcached import cached, cachedmethod
import sys

class BWMemberProperty(property):
    pass

NOT_FOUND = type(None)

class BWMember(BWObject):
    ro = False
    default = AttributeError
    builder = None
    optional = False

    def __init__(self, *isa, **_kw):
        self.isa = isa
        self.init(**_kw)

    def init(self, ro=None, default=NOT_FOUND,
                   optional=None, builder=None):
        if ro is not None:
            self.ro = ro
        if default is not NOT_FOUND:
            self.default = default
        if builder is not None:
            self.builder = builder
        if optional is not None:
            self.optional = optional
        elif default is not NOT_FOUND or builder:
            self.optional = True

    def __bindclass__(self, cls, name):
        p = BWMemberProperty(self.get_reader(cls, name),
                             self.get_writer(cls, name),
                             self.get_deleter(cls, name))
        p.__initobj__ = self.__initobj__
        p.__name__ = name
        members = cls.__dict__.get('__bwmembers__')
        if members is None:
            members = []
            for base in cls.__bases__:
                members.extend(getattr(base, '__bwmembers__', ()))
            members = tuple(members)
        cls.__bwmembers__ = (p,) + members
        if not self.optional:
            required = cls.__dict__.get('__required__')
            if required is None:
                required = []
                for base in cls.__bases__:
                    required.extend(getattr(base, '__required__', ()))
                required = tuple(required)
            required += (name,)
            cls.__required__ = sorted(required)
        return p

    @cachedmethod
    def checkset(self):
        return self.build_checker(*self.isa)

    def build_checker(self, *isa):
        src = ['def checker(_s, _n, _v):']
        lv = dict(NOT_FOUND=NOT_FOUND)
        def op(src, lv, indent):
            src.append(indent + 'return True')
        self.build_checker_src(isa, op, src, lv)
        src = '\n'.join(src)
        #print >>sys.stderr, src
        #print >>sys.stderr, lv
        exec src in lv
        checker = lv.pop('checker')
        checker.__src__ = src
        checker.__isa__ = isa
        return checker

    def build_checker_src(self, isa, op, src, lv, indent='    '):
        for check in isa:
            if check is None:
                src.append(indent + 'if _v is None:')
                op(src, lv, indent + '    ')
            elif isinstance(check, type):
                name = check.__name__
                lv[name] = check
                src.append(indent + 'if isinstance(_v, %s):' % name)
                op(src, lv, indent + '    ')
            elif callable(check):
                name = 'cb_%d' % len(lv)
                lv[name] = check
                if hasattr(check, '__converter__'):
                    allowed = getattr(check, '__from__', ())
                    def fn_op(src, lv, indent):
                        src.append(indent + 'try:')
                        src.append(indent + '    _nv = %s(_s, _n, _v)' % name)
                        src.append(indent + '    if _nv is not NOT_FOUND:')
                        src.append(indent + '        return (_nv,)')
                        src.append(indent + 'except:')
                        src.append(indent + '    pass')
                    tc = getattr(check, '__type__', None)
                    if tc is not None:
                        tname = tc.__name__
                        lv[tname] = tc
                        src.append(indent +
                                   'if isinstance(_v, %s):' % tname)
                        src.append(indent + '    return True')
                        src.append(indent + 'else:')
                        sindent = indent + '    '
                    else:
                        sindent = indent
                    if allowed:
                        self.build_checker_src(allowed, fn_op, src, lv, sindent)
                    else:
                        fn_op(src, lv, indent)
                else:
                    src.append(indent + 'if %s(_v, _s):' % name)
                    op(src, lv, indent + '    ')
            else:
                name = 'v_%d' % len(lv)
                lv[name] = check
                src.append(indent + 'if %s == _v:' % name)
                op(src, lv, indent + '    ')
        return src, lv

    def __initobj__(self, obj, name, value, NOT_FOUND=NOT_FOUND):
        res = self.checkset(name, value)
        while type(res) is tuple:
            value = res[0]
            res = self.checkset(name, value)
        if res:
            obj.__dict__[name,] = value
        else:
            raise TypeError('%s (%r) must be one of: (%s)'
                            % (name, value, ', '.join(map(repr, self.isa))))

    def get_reader(self, cls, name, NOT_FOUND=NOT_FOUND):
        def reader(o):
            obj = o.__dict__.get((name,), NOT_FOUND)
            if obj is NOT_FOUND:
                default = self.default
                if isinstance(default, type):
                    if issubclass(default, Exception):
                        obj = default
                    else:
                        obj = default()
                elif callable(default):
                    obj = default(o, name)
                else:
                    obj = default
                builder = self.builder
                if builder:
                    fn = getattr(o, builder, None)
                    if fn is None:
                        raise TypeError('%r has no builder method %r' %
                                        (self, builder))
                    else:
                        obj = fn(default)
                if isinstance(obj, type) and issubclass(obj, Exception):
                    raise obj(name)
                self.__initobj__(o, name, obj)
            return obj
        return reader

    def get_writer(self, cls, name):
        if self.ro:
            return None
        else:
            return lambda o, v: self.__initobj__(o, name, v)

    def get_deleter(self, cls, name):
        if self.ro:
            return None
        else:
            # XXX: Should we just default to no-op instead when not found?
            # I would personally prefer to make "del x.y" an error-free op.
            return lambda o: self.__dict__.pop((name,))

def member(*_args, **_kw):
    return BWMember(*_args, **_kw)

def into(_converter, *_isa):
    def converter(_s, _n, _v):
        return _converter(_v)
    name = '<converter'
    converter.__converter__ = True
    if isinstance(_converter, type):
        name += ' to %r' % _converter
        converter.__type__ = _converter
    else:
        name += ' fn %r' % _converter
    if _isa:
        converter.__from__ = _isa
        name += ' from %r' % _isa
    converter.__name__ = name + '>'
    return converter




'''
setup -- Setup script for bullwinkle
'''

from bullwinkle.__version__ import *
from setuptools import setup

if not VERSION.label:
    raise RuntimeError("Change log not written for %s" % (VERSION,))

setup(
    name = 'bullwinkle',
    version = VERSION,
    author = AUTHOR,
    author_email = AUTHOR_EMAIL,
    description = 'Python tools inspired by Perl::Moose',
    licesne = LICENSE,
    keywords = 'Moose OOP super',
    url = 'http://code.google.com/p/bullwinkle',
    packages = ['bullwinkle'],
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python',
        'Topic :: Utilities',
    ]
)




'''
__changelog__ -- contains the project changelog in object form.
'''

from bwversion import Version, ChangeLog

CHANGELOG = ChangeLog(
    Version('0.1', 'Initial pre-release at PyOhio'),
    Version('0.2.5',
        'Added throwables',
        'Added changelog file',
        'Modified version to be more shell friendly',
        'Made __version__ more pydoc compatible',
        ),
    Version('0.2.6',
        'Added throw() and catch() as simple functions',
        ),
    Version('0.3',
        'Added type conversion support for members',
        'Added default and built values for members',
        'Added required members',
        'Added TC global',
        )
)

if __name__ == '__main__':
    print CHANGELOG

